import pytest
import numpy as np

def test_analytical_errors(helpers):
    passing = []
    for (lam, tend, loc, lay) in zip(
            helpers.exps['lam_profile'],
            helpers.exps['extensive_tendency'],
            helpers.exps['local_wmt'],
            helpers.exps['layer_wmt']
        ):
        # pick discretizations with no shared multiples so levels do not coincide
        wmt = helpers.idealized_transformations(tend, lam, Nz=5**6, Nlam=2**5)
        err = helpers.mean_absolute_relative_errors(wmt, loc, lay)[0]
        passing.append(err < 1.e-3)
    assert np.all(passing)

def test_analytical_error_convergence(helpers):
    passing = []
    for (lam, tend, loc, lay) in zip(
            helpers.exps['lam_profile'],
            helpers.exps['extensive_tendency'],
            helpers.exps['local_wmt'],
            helpers.exps['layer_wmt']
        ):
        wmt_lo =     helpers.idealized_transformations(tend, lam, Nz=5**4, Nlam=2**4)
        wmt_hi_z =   helpers.idealized_transformations(tend, lam, Nz=5**6, Nlam=2**4)
        wmt_hi_lam = helpers.idealized_transformations(tend, lam, Nz=5**4, Nlam=2**6)
        err_lo =     helpers.mean_absolute_relative_errors(wmt_lo,     loc, lay)
        err_hi_z =   helpers.mean_absolute_relative_errors(wmt_hi_z,   loc, lay)
        err_hi_lam = helpers.mean_absolute_relative_errors(wmt_hi_lam, loc, lay)
        passing.append(
            (err_hi_z[0]   <  err_lo[0]) and
            (err_hi_z[1]   <  err_lo[1]) and
            (err_hi_lam[1] <  err_lo[1])
        )
    assert np.all(passing)



import numpy as np
import xarray as xr
import xgcm
import gsw
import warnings

class WaterMass:
    """
    Stores water mass characteristics and supports methods for analyzing water masses on a numerical grid.
    """
    def __init__(
        self,
        grid,
        t_name="thetao",
        s_name="so",
        h_name="thkcello",
        teos10=True,
        cp=3992.0,
        rho_ref=1035.0,
        t_var="conservative",
        s_var="absolute",
        ):
        """
        Create a new WaterMass object from an input xgcm.Grid instance.

        Parameters
        ----------
        grid: xgcm.Grid
            Contains information about ocean model grid coordinates, metrics, and data variables.
        t_name: str (default: "thetao")
            Name of conservative temperature variable [in degrees Celsius] in ds.
        s_name: str (default: "so")
            Name of absolute salinity variable [in g/kg] in ds.
        h_name: str (default: "thkcello")
            Name of thickness variable [in m] in ds.
        teos10 : boolean (default: True)
            Get expansion/contraction coefficients from the Thermodynamic Equation Of Seawater - 2010 (TEOS-10),
            unless "alpha" and "beta" variables already present in `grid._ds`.
        cp: float (default: 3992.0)
            Value of specific heat capacity.
        rho_ref: float (default: 1035.0)
            Value of reference potential density, assuming Boussinesq approximation.
        t_var: str ("absolute" or "practical")
            Does variable `t_name` represent "conservative", "potential", or "in-situ" temperature?
        s_var: str ("absolute" or "practical")
            Does variable `s_name` represent "absolute" or "practical" salinity?
        """
        # Grid copy
        self.grid = xgcm.Grid(
            grid._ds.copy(),
            coords={
                **{ax:grid.axes[ax].coords for ax in grid.axes.keys()},
            },
            metrics={k:vv.name for (k,v) in grid._metrics.items() for vv in v},
            boundary={
                **{ax:grid.axes[ax]._boundary for ax in grid.axes.keys()},
            },
            autoparse_metadata=False
        )
        self.t_name = t_name
        self.t_var = t_var
        self.s_name = s_name
        self.s_var = s_var
        self.h_name = h_name
        self.teos10 = teos10
        self.cp = cp
        self.rho_ref = rho_ref
        
        if "Z_metrics" in vars(self.grid):
            pass
        elif self.h_name in self.grid._ds:
            # Conservatively interpolate thickness to cell interfaces, needed to
            # estimate depth of layer centers and compute surface flux divergences
            Z_center_extended = np.concatenate((
                self.grid._ds[self.grid.axes['Z'].coords['outer']][np.array([0])].values,
                self.grid._ds[self.grid.axes['Z'].coords['center']].values,
                self.grid._ds[self.grid.axes['Z'].coords['outer']][np.array([-1])].values
            ))
            with warnings.catch_warnings():
                warnings.simplefilter(action='ignore', category=FutureWarning)
                if "time" in self.grid._ds:
                    time_attrs = self.grid._ds.time.attrs.copy()
                self.grid._ds[f'{h_name}_i'] = self.grid.transform(
                    self.grid._ds[h_name].fillna(0.),
                    "Z",
                    Z_center_extended,
                    method="conservative",
                ).assign_coords({
                    self.grid.axes['Z'].coords['outer']:
                    self.grid._ds[self.grid.axes['Z'].coords['outer']].values
                })
            setattr(self.grid, "Z_metrics", {
                "center": self.grid._ds[self.h_name],
                "outer": self.grid._ds[f'{self.h_name}_i']
            })
            if "time" in self.grid._ds:
                self.grid._ds.time.attrs = time_attrs # For some reason these are not preserved by default
        elif "Z" not in self.grid.axes:
            self.grid._ds["z_l"] = xr.DataArray([0.5], dims=("z_l",))
            self.grid._ds["z_i"] = xr.DataArray([0, 1.], dims=("z_i",))
            self.grid._ds[f"{self.h_name}"] = xr.DataArray([1], dims=("z_l",))
            self.grid._ds[f"{self.h_name}_i"] = xr.DataArray([0.5, 0.5], dims=("z_i",))
            coords_2d = {ax:self.grid.axes[ax].coords for ax in self.grid.axes.keys()}
            coords_2d["Z"] = {"center": "z_l", "outer": "z_i"}
            metrics_2d = {k:vv.name for (k,v) in self.grid._metrics.items() for vv in v}
            boundary_2d = {ax:self.grid.axes[ax]._boundary for ax in self.grid.axes.keys()}
            boundary_2d["Z"] = "extend"
            self.grid = xgcm.Grid(
                self.grid._ds,
                coords=coords_2d,
                metrics=metrics_2d,
                boundary=boundary_2d,
                autoparse_metadata=False
            )
            setattr(self.grid, "Z_metrics", {
                "center": self.grid._ds[f"{self.h_name}"],
                "outer": self.grid._ds[f"{self.h_name}_i"]
            })
        self.grid._ds['z'] = (
            -self.grid.cumsum(self.grid.Z_metrics["outer"], "Z")
        ).chunk({self.grid.axes["Z"].coords["center"]: -1})
        self.grid._ds['z_interface'] = xr.where(
            self.grid.axes["Z"].coords["outer"] != self.grid.axes["Z"].coords["outer"][0],
            -self.grid.cumsum(self.grid.Z_metrics["center"], "Z", to="outer"),
            0.
        ).chunk({self.grid.axes["Z"].coords["outer"]: -1})
        
    def get_density(self, density_name="rho"):
        """
        Derive density variables from layer temperature, salinity, and thickness,
        and add them to the dataset (if not already present).
        Uses the TEOS10 algorithm from the `gsw` package by default, unless "alpha"
        and "beta" variables are already provided in `self.grid._ds`.

        Parameters
        ----------
        density_name: str (default: "rho")
            Name of density variable. Supported density variables are: 
            "rho" (in-situ), "sigma0", "sigma1", "sigma2", "sigma3", "sigma4"
            (corresponding to functions of the same name in the `gsw` package).

        Returns
        -------
        xr.DataArray
        """
        
        if self.t_name not in self.grid._ds:
            raise ValueError(f"ds must include temperature variable\
            defined by kwarg t_name (default: {self.t_name}).")
        if self.s_name not in self.grid._ds:
            raise ValueError(f"ds must include salinity variable\
            defined by kwarg s_name (default: {self.s_name}).")
        if self.h_name not in self.grid._ds:
            raise ValueError(f"ds must include thickness variable\
            defined by kwarg h_name (default: {self.h_name}).")
        
        if (
            "alpha" not in self.grid._ds or "beta" not in self.grid._ds or self.teos10
        ) and "p" not in self.grid._ds.data_vars:
            self.grid._ds['p'] = xr.apply_ufunc(
                gsw.p_from_z, self.grid._ds.z, self.grid._ds.lat, 0, 0, dask="parallelized"
            )

        if "sigma" in density_name:
            z_ref = density_name.replace("sigma", "")
            try:
                z_ref = -float(z_ref)*1000
            except:
                print("'density_name' is not of form 'sigmaX' where 'X' is a number.")
            
            p_ref = xr.apply_ufunc(
                gsw.p_from_z, z_ref, self.grid._ds.lat, 0, 0, dask="parallelized"
            )
        elif density_name == "rho":
            z_ref = self.grid._ds.z
            p_ref = self.grid._ds.p
        else:
            raise NameError(f"`density_name = {density_name}` is not a supported option.")
        
        # Prognostic temperature and salinity are, by default, interpreted as
        # conservative temperature and absolute salinity (following McDougall et al. 2021).
        if self.teos10 and "sa" not in self.grid._ds:
            if self.s_var == "absolute":
                self.grid._ds['sa'] = self.grid._ds[self.s_name]
            elif self.s_var == "practical":
                self.grid._ds['sa'] = xr.apply_ufunc(
                    gsw.SA_from_SP,
                    self.grid._ds[self.s_name],
                    self.grid._ds.p,
                    self.grid._ds.lon,
                    self.grid._ds.lat,
                    dask="parallelized",
                )
        if self.teos10 and "ct" not in self.grid._ds:
            if self.t_var == "conservative":
                self.grid._ds['ct'] = self.grid._ds[self.t_name]
            elif self.t_var == "potential":
                self.grid._ds['ct'] = xr.apply_ufunc(
                    gsw.CT_from_pt,
                    self.grid._ds.sa,
                    self.grid._ds[self.t_name],
                    dask="parallelized"
                )
            elif self.t_var == "in-situ":
                self.grid._ds['ct'] = xr.apply_ufunc(
                    gsw.CT_from_t,
                    self.grid._ds.sa,
                    self.grid._ds[self.t_name],
                    self.grid._ds.p,
                    dask="parallelized"
                )
        if not self.teos10 and ("sa" not in vars(self) or "ct" not in vars(self)):
            self.grid._ds['sa'] = self.grid._ds[self.s_name]
            self.grid._ds['ct'] = self.grid._ds[self.t_name]

        # Calculate thermal expansion coefficient alpha (1/K) at reference pressure
        if "alpha" not in self.grid._ds:
            self.grid._ds['alpha'] = xr.apply_ufunc(
                gsw.alpha, self.grid._ds.sa, self.grid._ds.ct, p_ref, dask="parallelized"
            )

        # Calculate the haline contraction coefficient beta (kg/g) at reference pressure
        if "beta" not in self.grid._ds:
            self.grid._ds['beta'] = xr.apply_ufunc(
                gsw.beta, self.grid._ds.sa, self.grid._ds.ct, p_ref, dask="parallelized"
            )

        # Calculate potential density (kg/m^3)
        if density_name not in self.grid._ds:
            if density_name == "rho":
                self.grid._ds[density_name] = xr.apply_ufunc(
                    getattr(gsw, density_name),
                    self.grid._ds.sa,
                    self.grid._ds.ct,
                    self.grid._ds.p,
                    dask="parallelized"
                ).rename(density_name)
            
            elif "sigma" in density_name:
                self.grid._ds[density_name] = xr.apply_ufunc(
                    getattr(gsw, density_name),
                    self.grid._ds.sa,
                    self.grid._ds.ct,
                    dask="parallelized"
                ).rename(density_name)

        return self.grid._ds[density_name]

    def get_outcrop_lev(self, position="center", incrop=False, min_thickness=1e-6):
        """
        Return the first vertical level (starting from the relevant boundary)
        whose thickness exceeds `min_thickness`.
    
        Assumes the native Z index increases downward (surface -> bottom).
    
        Parameters
        ----------
        position : {"center", "outer"}
            Vertical grid position for the coordinate/metric.
        incrop : bool
            False (default): search from sea surface downward ("outcrop").
            True: search from seafloor upward ("incrop").
        min_thickness : float
            Default: 1e-6. Minimum layer thickness required to count as the first "real" cell.
    
        Returns
        -------
        xr.DataArray
            Vertical coordinate value, broadcast across other dims, masked where no
            layer exceeds `min_thickness`.
        """
        z_coord = self.grid.axes["Z"].coords[position]
        h = self.grid.Z_metrics[position]
    
        # Native order is surface -> bottom. For incrop, reverse to bottom -> surface.
        z_native = self.grid._ds[z_coord]
        z_order = z_native[::-1] if incrop else z_native
    
        h_ord = h.sel({z_coord: z_order})
    
        thick = h_ord > min_thickness
    
        # True only at the first thick cell in boundary->interior order
        first = (thick.cumsum(z_coord) == 1) & thick
    
        out = first.astype("int8").idxmax(z_coord)
    
        # Mask columns where no thick cell exists
        return out.where(thick.any(z_coord))

        
    def sel_outcrop_lev(self, da, incrop=False, min_thickness=1e-6, position="center", **kwargs):
        """
        Select `da` at the first vertical level (starting from the relevant boundary)
        whose thickness exceeds `min_thickness`.
    
        Assumes the native Z index increases downward (surface -> bottom).
    
        Parameters
        ----------
        da : xr.DataArray
            DataArray to select from. Must have the same dims as the thickness metric
            after applying `**kwargs`.
        incrop : bool
            Default: False. If True, selects the first thick level from the bottom upward.
        min_thickness : float
            Minimum layer thickness required to count as the first "real" cell.
        position : {"center", "outer"}
            Vertical grid position for the coordinate/metric.
        **kwargs : dict
            Passed to `.sel(**kwargs)` on the thickness metric (and used to validate dims).
    
        Returns
        -------
        xr.DataArray
            `da` selected at the diagnosed level, masked where no level exceeds `min_thickness`.
        """
        z_coord = self.grid.axes["Z"].coords[position]
    
        # Thickness metric, subset as requested
        h = self.grid.Z_metrics[position].sel(**kwargs)
    
        missing_dims = set(h.dims) - set(da.dims)
        if missing_dims:
            raise ValueError(
                f"`da` is missing required dimensions {missing_dims}. "
                f"`da.dims={da.dims}`, `h.dims={h.dims}`"
            )

        # Order so the relevant boundary is first along z (native is surface->bottom)
        z_native = self.grid._ds[z_coord]
        z_order = z_native[::-1] if incrop else z_native
    
        h_ord = h.sel({z_coord: z_order})
        da_ord = da.sel({z_coord: z_order})
    
        thick = h_ord > min_thickness
    
        # First thick cell in this boundary->interior ordering
        first = (thick.cumsum(z_coord) == 1) & thick
        lev = first.astype("int8").idxmax(z_coord)
    
        has_thick = thick.any(z_coord)
    
        # Select from the *original* `da` (order doesn’t matter once we have coord values)
        return da.sel({z_coord: lev}).where(has_thick)
    
    def expand_surface_array_vertically(self, da_surf, target_position="outer"):
        """
        Expand surface xr.DataArray (with no "Z"-dimension coordinate) in the vertical,
        filling with zeros in all layers except the one that outcrops.
        
        Parameters
        ----------
        da_surf: xarray.DataArray
            Variable that is to be expanded in the vertical.
        position : str
            Position of the desired vertical coordinate in the `self.grid` instance of `xgcm.Grid`.
            Default: "outer". Other supported option is "center".
        """
        z_coord = self.grid.axes['Z'].coords[target_position]
        return (
            da_surf.expand_dims({z_coord: self.grid._ds[z_coord]})
            .where(
                self.grid._ds[z_coord] ==
                self.get_outcrop_lev(position=target_position),
                0.
            )
        )

    def infer_bins(self, da, percentiles=[0., 1.], nbins=100, surface=False):
        """
        Specify bins based on the distribution of `da`, excluding outliers.
        
        Parameters
        ----------
        da: xarray.DataArray
            Variable used to determine bins.
        percentiles: list
            List of length 2 containing the upper and lower percentiles to bound the array of bins.
            Default: [0., 1.], i.e. min and max.
        nbins: int
            Number of bins. Default: 100.
        surface: bool
            Default: False. If True, compute percentiles only from the outcropping layer of `da`.
        """
        if surface:
            da=self.sel_outcrop_lev(da)
        if percentiles != [0., 1.]:
            vmin, vmax = da.quantile(percentiles, dim=da.dims)
        else:
            vmin, vmax = da.min(), da.max()
        return np.linspace(vmin, vmax, nbins)

    def zonal_mean(self, da, oceanmask_name="wet"):
        """
        Compute area-weighted zonal mean (along `X` grid axis).
        
        Parameters
        ----------
        da: xarray.DataArray
            Data array to be averaged.
        oceanmask_name: str
            Name of ocean mask xr.DataArray in `self.grid._ds`. Default: "wet".
        """
        x_name = grid.axes['X'].coords['center']
        area = self.grid.get_metric(da, ['X', 'Y'])
        num = (da * area * self.grid._ds[landmask_name]).sum(dim=x_name)
        denom = (area * self.grid._ds[landmask_name]).sum(dim=x_name)
        return num / denom

def add_gridcoords(grid, coords, boundary):
    new_grid = xgcm.Grid(
        grid._ds,
        coords={
            **{ax:grid.axes[ax].coords for ax in grid.axes.keys()},
            **coords
        },
        metrics={k:vv.name for (k,v) in grid._metrics.items() for vv in v},
        boundary={
            **{ax:grid.axes[ax]._boundary for ax in grid.axes.keys()},
            **boundary
        },
        autoparse_metadata=False
    )
    if "Z_metrics" in vars(grid):
        new_grid.Z_metrics = grid.Z_metrics

    return new_grid


import pytest
import numpy as np
import xarray as xr
import xbudget
import xgcm
import xwmt

ds = xr.open_dataset("xwmb_test_data_Baltic_3d.20230830.nc", decode_timedelta=False).isel(time=0)
coords = {
    'X': {'center': 'xh', 'outer': 'xq'},
    'Y': {'center': 'yh', 'outer': 'yq'},
    'Z': {'center': 'zl', 'outer': 'zi'},
}
metrics = {
    ('X','Y'): "areacello",
}
grid = xgcm.Grid(ds, coords=coords, metrics=metrics, periodic=None, autoparse_metadata=False)

budgets_dict = xbudget.load_preset_budget(model="MOM6")
# The test data set does not include sea ice melt diagnostics
del budgets_dict["mass"]["rhs"]["sum"]["surface_exchange_flux"]["sum"]["sea_ice_melt"]
xbudget.collect_budgets(grid, budgets_dict)
simple_budgets = xbudget.aggregate(budgets_dict)

## Default parameters except that we group all processes together
kwargs = {'group_processes': True}

# heat
def test_functional_3d_theta():
    answer_dict = {
        "xgcm": np.array([5.89988332e+08, 1.71073947e+08, 1.60436210e+09, 6.14312533e+08]),
        "xhistogram": np.array([6.11109007e+08, 6.16747397e+08, 1.16849868e+09, 4.80601100e+08])
    }
    for method in ["xgcm", "xhistogram"]:
        wmt = xwmt.WaterMassTransformations(grid, simple_budgets, method=method)
        total_wmt = wmt.integrate_transformations(
            "heat",
            bins = np.linspace(0., 4., 5),
            **kwargs
            )['material_transformation']
        assert np.all(np.isclose(
            total_wmt.values,
            answer_dict[method]
        ))


# salt
def test_functional_3d_salt():
    answer_dict = {
        "xgcm": np.array([-6.92695360e+07, -2.92969936e+08, -5.65359683e+07,  1.31220262e+08]),
        "xhistogram": np.array([-7.09139741e+07, -2.94351853e+08, -2.40161246e+07,  6.65517384e+07])
    }
    for method in ["xgcm", "xhistogram"]:
        wmt = xwmt.WaterMassTransformations(grid, simple_budgets, method=method)
        total_wmt = wmt.integrate_transformations(
            "salt",
            bins = np.linspace(5., 9., 5),
            **kwargs
            )['material_transformation']
        assert np.all(np.isclose(
            total_wmt.values,
            answer_dict[method]
        ))
# sigma2
def test_functional_3d_sigma2():
    answer_dict = {
        "xgcm": np.array([-3.89013506e+08,  1.11459836e+08,  3.97737451e+07,  7.12295765e+06]),
        "xhistogram": np.array([-3.76907156e+08,  9.13575231e+07,  4.22416664e+07,  6.72588395e+06])
    }
    for method in ["xgcm", "xhistogram"]:
        wmt = xwmt.WaterMassTransformations(grid, simple_budgets, method=method)
        total_wmt = wmt.integrate_transformations(
            "sigma2",
            bins=np.linspace(15., 19., 5),
            **kwargs
            )['material_transformation']
        assert np.all(np.isclose(
            total_wmt.values,
            answer_dict[method]
        ))



import pytest
import os
import numpy as np
import xarray as xr
import xgcm
import xwmt
import urllib.request

fname = 'xwmb_test_data_Baltic_3d.20230830.nc'
ftp_path = 'ftp://ftp.gfdl.noaa.gov/perm/John.Krasting/xwmt/'
if not os.path.isfile(fname):
    print(f'Downloading test dataset from {ftp_path}{fname}')
    urllib.request.urlretrieve(f'{ftp_path}{fname}', fname)

class Helpers:

    def __init__(self):
        pass

    def idealized_transformations(self, extensive_tendency, lam_profile, Nz=1e3, Nlam=8):
        bins = np.linspace(0., 1., int(Nlam)+1)
        
        ds = xr.Dataset()
        ds = ds.assign_coords({
            'z_i': xr.DataArray(np.linspace(0., 1., int(Nz)+1), dims=("z_i",)),
            'z_l': xr.DataArray(np.linspace(1. /Nz, 1. - 1. /Nz, int(Nz)), dims=("z_l",)),
        })
        ds = ds.assign_coords({'dz': xr.DataArray(np.diff(ds.z_i.values), dims=("z_l",))})
        extensive_tendency_method = getattr(self, extensive_tendency)
        ds['tendency_name'] = xr.DataArray(extensive_tendency_method(ds.z_i.values), coords=(ds.z_l,))*ds.dz
        lam_profile_method = getattr(self, lam_profile)
        ds['temperature'] = xr.DataArray(lam_profile_method(ds.z_l.values), coords=(ds.z_l,))

        # expand to horizontal dimension and add grid metrics
        ds = ds.expand_dims(dim=('x', 'y')).assign_coords({'x':xr.DataArray([1.], dims=('x',)), 'y':xr.DataArray([1.], dims=('y',))})
        ds = ds.assign_coords({'rA': xr.DataArray([[1.]], dims=('x','y',))})

        metrics = {
            ('X', 'Y'): ['rA'] # Areas
        }
        coords = {
            'X': {'center': 'x',},
            'Y': {'center': 'y',},
            'Z': {'center': 'z_l', 'outer': 'z_i'},
        }
        grid = xgcm.Grid(ds, coords=coords, metrics=metrics, periodic=False, autoparse_metadata=False)
        
        budget_dict = {
            'mass': {
                'lambda': None,
                'thickness': 'dz',
                'rhs': {},
                'lhs': {}
            },
            'heat': {
                'lambda': 'temperature',
                'lhs': {'tendency':'tendency_name'},
                'rhs': {}
            },
            'salt': {
                'lambda': None,
                'rhs': {},
                'lhs': {}
            }
        }

        wmt = xwmt.WaterMassTransformations(grid, budget_dict, cp=1., rho_ref=1., method="xgcm")
        T = wmt.integrate_transformations("heat", bins=bins, sum_components=False)
        T = T.assign_coords({'temperature_i_target': xr.DataArray(bins, dims=("temperature_i_target",))})
        return T
    
    def mean_absolute_relative_errors(self, wmt_xwmt, wmt_local_exact, wmt_layer_exact):
        def absolute_relative_errors(wmt, wmt_ref):
            return np.abs((wmt - wmt_ref)/wmt_ref).where(np.abs(wmt_ref)>1.e-4).mean(skipna=True).values
        wmt_local_exact_method = getattr(self, wmt_local_exact)
        wmt_layer_exact_method = getattr(self, wmt_layer_exact)
        return (
            absolute_relative_errors(
                wmt_xwmt['tendency'],
                wmt_layer_exact_method(wmt_xwmt.temperature_i_target.values)
            ),
            absolute_relative_errors(
                wmt_xwmt['tendency'],
                wmt_local_exact_method(wmt_xwmt.temperature_l_target.values)
            )
        )

    # Extensive (layer-integrated) analytical tendency profiles
    def diffusive_extensive_tendency(self, z_i):
        def f(z):
            return -np.cos(2. *np.pi*z)/(2. *np.pi)
        return np.diff(f(z_i))/np.diff(z_i)

    def constant_extensive_tendency(self, z_i):
        return np.diff(z_i)/np.diff(z_i)

    def constant_plus_diffusion_tendency(self, z_i):
        return self.diffusive_extensive_tendency(z_i) + self.constant_extensive_tendency(z_i)

    def differential_heating_layer(self, z_i):
        sign = 2. *np.float64(z_i>0.5)-1
        out = np.diff(sign*z_i)/np.diff(z_i)
        out[(z_i[:-1]<=0.5)&(0.5<=z_i[1:])] = 0.
        return out

    # Stratification profiles
    def lam_const_dlamdz(self, z):
        return z

    def lam_linear_dlamdz(self, z):
        return z**2

    def lam_overturning_dlamdz(self, z):
        return 1 - (2*z - 1)**2

    def lam_vanishing_dlamdz(self, z):
        sign = 2*np.float64(z>0.5)-1
        return (1 + sign*(2*z - 1)**2)/2.

    # Analytical point-wise water mass transformations
    def constant_plus_diffusion_local_wmt_dlamdz_constant(self, lam):
        return -(np.sin(2*np.pi*lam) + 1.)

    def constant_plus_diffusion_local_wmt_dlamdz_linear(self, lam):
        return -(np.sin(2*np.pi*np.sqrt(lam)) + 1.)/(2*np.sqrt(lam))

    def constant_plus_diffusion_local_wmt_dlamdz_overturning(self, lam):
        return -1/(2*np.sqrt(1-lam))

    def differential_heating_local_wmt_dlamdz_vanishing(self, lam):
        sign = 2*np.float64(lam>0.5)-1
        return -sign/(2*np.sqrt(sign*(2*lam-1)))

    # Analytical layer-averaged water mass transformations
    def constant_plus_diffusion_layer_wmt_dlamdz_constant(self, lam_bins):
        def f(lam):
            return -np.cos(2*np.pi*lam)/(2*np.pi) + lam
        return -np.diff(f(lam_bins))/np.diff(lam_bins)

    def constant_plus_diffusion_layer_wmt_dlamdz_linear(self, lam_bins):
        def f(lam):
            return -np.cos(2*np.pi*np.sqrt(lam))/(2*np.pi) + np.sqrt(lam)
        return -np.diff(f(lam_bins))/np.diff(lam_bins)

    def constant_plus_diffusion_layer_wmt_dlamdz_overturning(self, lam_bins):
        def f(lam):
            return 1 - np.sqrt(1 - lam)
        return -np.diff(f(lam_bins))/np.diff(lam_bins)

    def differential_heating_layer_wmt_dlamdz_vanishing(self, lam_bins):
        def f(lam):
            sign = 2*np.float64(lam>0.5)-1
            u = 1 + sign*np.sqrt(sign*(2*lam-1))
            return sign*0.5*u
        out = -np.diff(f(lam_bins))/np.diff(lam_bins)
        out[(lam_bins[:-1]<=0.5)&(0.5<=lam_bins[1:])] = np.nan
        return out

    exps = {
        "extensive_tendency": [
            "constant_plus_diffusion_tendency",
            "constant_plus_diffusion_tendency",
            "constant_plus_diffusion_tendency",
            "differential_heating_layer"
        ],
        "lam_profile": [
            "lam_const_dlamdz",
            "lam_linear_dlamdz",
            "lam_overturning_dlamdz",
            "lam_vanishing_dlamdz"
        ],
        "local_wmt": [
            "constant_plus_diffusion_local_wmt_dlamdz_constant",
            "constant_plus_diffusion_local_wmt_dlamdz_linear",
            "constant_plus_diffusion_local_wmt_dlamdz_overturning",
            "differential_heating_local_wmt_dlamdz_vanishing"
        ],
        "layer_wmt":[
            "constant_plus_diffusion_layer_wmt_dlamdz_constant",
            "constant_plus_diffusion_layer_wmt_dlamdz_linear",
            "constant_plus_diffusion_layer_wmt_dlamdz_overturning",
            "differential_heating_layer_wmt_dlamdz_vanishing"
        ]
    }

@pytest.fixture
def helpers():
    return Helpers()



import warnings

import gsw
import numpy as np
import xarray as xr
import xgcm

def hlamdot_from_Jlam(grid, Jlam, dim):
    """
    Calculation of hlamdot (cell-depth integral of scalar tendency)
    from interfacial fluxes
    """
    # For convergence, need to reverse the sign
    dJlam = -grid.diff(Jlam, dim)
    if "Z_metrics" in list(vars(grid)):
        h = grid.Z_metrics["center"]
        h = h.where(h!=0.)
    else:
        h = grid.get_metric(dJlam, "Z")
    lamdot = dJlam/h
    hlamdot = h.fillna(0.)*lamdot.fillna(0.)
    return hlamdot

def calc_hlamdot_tendency(grid, datadict):
    """
    Wrapper functions to determine h times lambda_dot (vertically extensive tendency)
    """

    if "layer_integrated_tendency" in datadict:
        return datadict["layer_integrated_tendency"]
    
    elif "interfacial_flux" in datadict:
        return hlamdot_from_Jlam(
            grid,
            datadict["interfacial_flux"],
            "Z"
        )


import copy
import numpy as np
import xarray as xr
from xhistogram.xarray import histogram
import warnings

from xbudget import flatten_lol
from xwmt.wm import WaterMass
from xwmt.compute import calc_hlamdot_tendency

class WaterMassTransformations(WaterMass):
    """
    An extension of the WaterMass class that includes methods for WaterMass transformation analysis.
    """
    def __init__(
        self,
        grid,
        xbudget_dict,
        mask=None,
        teos10=True,
        cp=3992.0,
        rho_ref=1035.0,
        t_var="conservative",
        s_var="absolute",
        method="default",
        rebin=False,
        ):
        """
        Create a new WaterMassTransformation object from an input xgcm.Grid and xbudget dictionary.

        Parameters
        ----------
        grid : xgcm.Grid
            Contains information about ocean model grid coordinates, metrics, and data variables.
        xbudget_dict : dict
            Nested dictionary containing information about lambda and tendency variable names.
            See the `xbudget` package documentation (https://github.com/hdrake/xbudget) for how 
            this dictionary should be structured. In particular, the `xbudget/conventions`
            directory contains example `.yaml` files that can be read in as preset dictionaries.
            The `MOM6.yaml` file provides a comprehensive description of the mass, heat, and salt
            budgets in MOM6 (github.com/hdrake/xbudget/blob/main/xbudget/conventions/MOM6.yaml).
        mask : xr.DataArray (default: None)
            Boolean region mask (with same X and Y grid dimensions as `grid._ds` variables).
            If None, generate an all-True mask for domain-wide calculations.
        teos10 : bool (default: True)
            Use Thermodynamic Equation Of Seawater - 2010 (TEOS-10). True by default.
        cp : float (default: 3992.0, the MOM6 default value)
            Value of specific heat capacity.
        rho_ref : float (default: 1035.0, the MOM6 default value)
            Value of reference potential density. Note: WaterMass is assumed to be Boussinesq.
        t_var: str ("conservative", "potential", or "in-situ")
            Does variable `t_name` represent "conservative", "potential", or "in-situ" temperature?
        s_var: str ("absolute" or "practical")
            Does variable `s_name` represent "absolute" or "practical" salinity?
        method : str (default: "default")
            Method used for vertical transformations.
            Supported options: "default", "xhistogram", "xgcm".
            If "default", use "xhistogram" for area-integrated calculations (`integrate=True`)
            or "xgcm" for column-wise calculations (`integrate=False`) for efficiency.
            The other options force the use of a specific method, perhaps at the cost of efficiency.
        rebin : bool (default: False)
            Set to True to force a transformation into the target coordinates, even if these
            coordinates already exist in the `grid` data structure.
        """
        
        self.method = method
        self.rebin = rebin
        self.tracer_dict = {}
        tracers = [k for k in xbudget_dict.keys() if k!="mass"]
        for tracer in tracers:
            if "lambda" in xbudget_dict[tracer]:
                self.tracer_dict[tracer] = xbudget_dict[tracer]["lambda"]
            elif "surface_lambda" in xbudget_dict[tracer]:
                self.tracer_dict[tracer] = xbudget_dict[tracer]["surface_lambda"]
            else:
                raise ValueError(f"""xbudget_dict must contain element `["{tracer}"]["lambda"]` `["{tracer}"]["surface_lambda"]`""")

        kwargs = {}
        contains_thickness = False
        if "mass" in xbudget_dict:
            if "thickness" in xbudget_dict["mass"]:
                kwargs["h_name"] = xbudget_dict["mass"]["thickness"]
                contains_thickness = True
        elif not(contains_thickness):
            raise ValueError("""xbudget_dict must contain element `["mass"]["thickness"]`""")

        super().__init__(
            grid,
            t_name=self.tracer_dict.get("heat"), # defaults to None if not available
            s_name=self.tracer_dict.get("salt"), # defaults to None if not available
            teos10=teos10,
            cp=cp,
            rho_ref=rho_ref,
            t_var=t_var,
            s_var=s_var,
            **kwargs
        )

        self.lambdas_dict = self.tracer_dict
        if all(k in self.tracer_dict for k in ["heat", "salt"]):
            self.lambdas_dict = {**self.lambdas_dict, **{"density": ["sigma0", "sigma1", "sigma2", "sigma3", "sigma4"]}}
        
        self.xbudget_dict = copy.deepcopy(xbudget_dict)
        for (term, bdict) in self.xbudget_dict.items():
            setattr(self, f"processes_{term}_dict", {})
            for ptype, _processes in bdict.items():
                if ptype in ["lhs", "rhs"]:
                    getattr(self, f"processes_{term}_dict").update(_processes)

    def lambdas(self, lambda_key=None):
        """
        Return dictionary of desired lambdas.

        Parameters
        ----------
        lambda_key : str or list of str (default: None)
            Lambda(s) of interest. If None, return a list of all available lambdas.

        Returns
        -------
        list

        Example
        -------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt.lambdas()
        ['thetao', 'so', 'sigma0', 'sigma1', 'sigma2', 'sigma3', 'sigma4']
        """
        if lambda_key is None:
            return flatten_lol(self.lambdas_dict.values())
        else:
            return self.lambdas_dict.get(lambda_key, None)
        
    def get_lambda_var(self, lambda_name=None):
        """
        Search lambda dictionary for variable name corresponding to specified lambda

        Parameters
        ----------
        lambda_name : str (default: None)
            Name of lambda variable. Supported options: ["heat", "salt", "density"]
            If None, return None

        Returns
        -------
        str or list

        Example
        -------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt.get_lambda_var("heat")
        'thetao'
        
        >>> wmt.get_lambda_var("density")
        ['sigma0', 'sigma1', 'sigma2', 'sigma3', 'sigma4']
        
        See also
        --------
        self.lambdas, self.get_lambda_key
        """
        density_lambdas = self.lambdas_dict["density"] if "density" in self.lambdas_dict else []
        if lambda_name in self.lambdas_dict:
            return self.lambdas_dict[lambda_name]
        elif lambda_name in density_lambdas:
            return lambda_name
        else:
            return None
        
    def get_lambda_key(self, lambda_var):
        """
        Search lambda dictionary for lambda corresponding to variable name

        Parameters
        ----------
        lambda_var : str
            Variable name of lambda.

        Returns
        -------
        str

        Example
        -------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt.get_lambda_key("thetao")
        'heat'
        
        >>> wmt.get_lambda_var("sigma2")
        'density'

        See also
        --------
        self.lambdas, self.get_lambda_var
        """
        for k,v in self.lambdas_dict.items():
            if type(v) is str:
                if lambda_var == v:
                    return(k)
            elif type(v) is list:
                if lambda_var in v:
                    return(k)


    def process_names(self, tracer, term):
        """
        Get a tuple containing the names of variables for the 'tracer' and the 'process'-specific
        tendency corresponding to a general 'term' in the tendency equation.

        Parameters
        ----------
        tracer : str
            Supported options: ["heat", "salt"]
        term : str
            key for tendency variable in the xbudget_dict

        Returns
        -------
        names : tuple
            `(tracer_name, process)`

        Example
        -------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt.process_names("heat", "diffusion")
        ('thetao', 'opottempdiff')
        """
        if tracer in self.tracer_dict.keys():
            process = getattr(self, f"processes_{tracer}_dict").get(term, None)
        else:
            warnings.warn(f"Tracer {tracer} is not defined")
            return
        tracer_name = self.tracer_dict.get(tracer, None)
        return (tracer_name, process)

    def available_processes(self, available=True):
        """
        Get a list of all tendency processes that are both specified by `xbudget_dict` and available in
        the dataset.

        Parameters
        ----------
        available: bool
            Default True. If False, include processes that are not available in the dataset.

        Returns
        -------
        processes : list of str

        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> processes = wmt.available_processes()
        ['diffusion']
        """
        processes = set()
        for tracer in self.xbudget_dict.keys():
            processes |= getattr(self, f"processes_{tracer}_dict").keys()
        if available:
            _processes = []
            for process in processes:
                if all([
                    getattr(self, f"processes_{tracer}_dict").get(process, None) in self.grid._ds
                    for tracer in self.xbudget_dict.keys()
                    if getattr(self, f"processes_{tracer}_dict").get(process, None) is not None
                ]):
                    _processes.append(process)
            return _processes
        else:
            return processes

    def datadict(self, tracer, term):
        """
        Get a dictionary that organizes the variables and metadata necessary to evaluate tendency terms.

        Parameters
        ----------
        component : str
            Supported options: ["heat", "salt"]
        term : str
            key for tendency variable in the xbudget_dict

        Returns
        -------
        ddict : dict

        Example
        --------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> ddict = wmt.datadict("heat", "diffusion")
        """
        (tracer_name, process) = self.process_names(tracer, term)
        
        if process is None or process not in self.grid._ds:
            return

        tend_arr = self.grid._ds[process]
        
        # Multiply salt tendency by 1000 to convert to g/m^2/s
        if tracer=="salt":
            tend_arr = tend_arr*1000.

        scalar = self.grid._ds[tracer_name]
        if self.grid.axes['Z'].coords["center"] not in scalar.dims:
            scalar = self.expand_surface_array_vertically(scalar, target_position="center")
        
        n_zcoords = len([
            c for c in self.grid.axes['Z'].coords.values()
            if c in self.grid._ds[process].dims
        ])
        
        if n_zcoords == 0:
            tend_arr = self.expand_surface_array_vertically(tend_arr, target_position="center")
            tend_dict = {"layer_integrated_tendency": tend_arr}
        elif n_zcoords > 0:
            if (self.grid.axes['Z'].coords["outer"] in self.grid._ds[process].dims):
                tend_dict = {"interfacial_flux": tend_arr}
            elif (self.grid.axes['Z'].coords["center"] in self.grid._ds[process].dims):
                tend_dict = {"layer_integrated_tendency": tend_arr}
        
        return {"scalar": scalar, **tend_dict}

    def rho_tend(self, term):
        """
        Get density tendency 'term' from underlying heat and salt tendencies. 

        Parameters
        ----------
        term : str
            key for tendency variable in the xbudget_dict

        Returns
        -------
        rho_tend_heat, rho_tend_salt
            The two distinct components contributing to the overall density tendency.

        Example
        --------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt.get_density()
        >>> (rho_tend_heat, rho_tend_salt) = wmt.rho_tend("diffusion")
        """

        # Either heat or salt tendency/flux may not be used
        rho_tend_heat, rho_tend_salt = None, None

        datadict = self.datadict("heat", term)
        if datadict is not None:
            heat_tend = calc_hlamdot_tendency(self.grid, datadict)
            # Density tendency due to heat flux (kg/s/m^2)
            rho_tend_heat = -(self.grid._ds.alpha / self.cp) * heat_tend

        datadict = self.datadict("salt", term)
        if datadict is not None:
            salt_tend = calc_hlamdot_tendency(self.grid, datadict)
            # Density tendency due to salt/salinity (kg/s/m^2)
            rho_tend_salt = self.grid._ds.beta * salt_tend

        return rho_tend_heat, rho_tend_salt

    def calc_hlamdot_and_lambda(self, lambda_name, term):
        """
        Get layer-integrated extensive tracer tendencies (* kg/m^2/s) and corresponding scalar field of lambda
        
        Parameters
        ----------
        lambda_name : str
            Specifies lambda
        term : str
            key for tendency variable in the xbudget_dict
            
        Returns
        ----------
        hlamdot, lam : xr.DataArray, xr.DataArray

        Example
        --------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> (hlamdot, lam) = wmt.calc_hlamdot_and_lambda("heat", "diffusion")
        """
        
        lam_var = self.get_lambda_var(lambda_name)
        prebinned = all([
            (c in self.grid.axes['Z'].coords.values())
            for c in [f"{lam_var}_l", f"{lam_var}_i"]
        ])

        # Get layer-integrated potential temperature tendency
        # from tendency of heat (in W/m^2)
        if lambda_name == "heat":
            datadict = self.datadict("heat", term)
            if datadict is not None:
                hlamdot = calc_hlamdot_tendency(self.grid, datadict) / self.cp
                lam = datadict["scalar"] if not prebinned else self.grid._ds[f"{lam_var}_l"]

        # Get layer-integrated practical salinity tendency
        # from tendency of salt (in g/s/m^2)
        elif lambda_name == "salt":
            datadict = self.datadict("salt", term)
            if datadict is not None:
                hlamdot = calc_hlamdot_tendency(self.grid, datadict)
                lam = datadict["scalar"] if not prebinned else self.grid._ds[f"{lam_var}_l"]

        # Get layer-integrated potential density tendencies (in kg/s/m^2)
        # from heat and salt, lambda = density
        # Here we want to output 2 separate components of the transformation rates:
        # (1) transformation due to heat tend, (2) transformation due to salt tend
        elif lambda_name in ([] if self.lambdas("density") is None else self.lambdas("density")):
            lam = self.get_density(lambda_name)
            rhos = self.rho_tend(term)
            hlamdot = {}
            for idx, tend in enumerate(["heat", "salt"]):
                if rhos[idx] is not None:
                    hlamdot[tend] = rhos[idx]*self.rho_ref # Is this correct for non-boussinesq case?
                elif rhos[idx] is None:
                    hlamdot[tend] = rhos[idx]
                    
            if prebinned and not(self.rebin):
                lam = self.grid._ds[f"{lam_var}_l"]
            elif lam_var in self.grid._ds:
                lam = self.grid._ds[lam_var]
        
        elif lambda_name in self.tracer_dict.keys():
            datadict = self.datadict(lambda_name, term)
            if datadict is not None:
                hlamdot = calc_hlamdot_tendency(self.grid, datadict)
                lam = datadict["scalar"] if not prebinned else self.grid._ds[f"{lam_var}_l"]

        else:
            raise ValueError(f"{lambda_name} is not available in the xbudget_dict.")
        
        try:
            return hlamdot, lam
        
        except NameError:
            return None, None

    def transform_hlamdot_term(self, lambda_name, term, bins=None, mask=None, integrate=True):
        """
        Lazily compute extensive tendencies and transform them into lambda space
        along the vertical ("Z") dimension.
        
        Parameters
        ----------
        lambda_name : str
            Specifies lambda (e.g., 'heat', 'salt', 'sigma0', etc.). Use `lambdas()` for a list of available lambdas.
        term : str
            Specifies process term (e.g., 'boundary_forcing', 'vertical_diffusion', etc.). Use `self.available_processes()`
            to see a list of all available terms.
        bins : array like (default None)
            np.array with lambda values specifying the edges for each bin. If not specified, array will be automatically
            derived from the scalar field of lambda (e.g., temperature).
        mask : xr.DataArray (default: None)
            Boolean region mask (with same X and Y grid dimensions as `grid._ds` variables).
            If None, generate an all-True mask for domain-wide calculations.
        integrate : bool (default: True)
            Whether to integrate the result over ("X", "Y") dimensions.
        
        Returns
        -------
        transformed_term : xarray.DataArray
            Data array containing transformed tendencies.

        See also
        --------
        self.transformations_from_hlamdot, self.integrate_transformations
        """

        hlamdot, lam = self.calc_hlamdot_and_lambda(lambda_name, term)
        if hlamdot is None:
            return

        if self.method in ["default", "xhistogram"]:
            if integrate:
                dim = [
                    self.grid.axes['X'].coords['center'],
                    self.grid.axes['Y'].coords['center'],
                    self.grid.axes['Z'].coords['center']
                ]
            else:
                dim = [self.grid.axes['Z'].coords['center']]

        if mask is not None:
            if type(hlamdot) is dict:
                hlamdot = {
                    k:v.where(mask, 0.)
                    if v is not None else None
                    for k,v in hlamdot.items()
                }
            else:
                hlamdot = hlamdot.where(mask, 0.)

        if bins is None:
            bins = self.infer_bins(lam)
            
        # If lambda is already a vertical coordinate, no need to use the 3D lambda for transformations
        lam_var = self.get_lambda_var(lambda_name)
        prebinned = all([(c in self.grid.axes['Z'].coords.values()) for c in [f"{lam_var}_l", f"{lam_var}_i"]])
        if prebinned and not(self.rebin):
            onedimensional_target = True
            lam = lam.rename({lam.name: lam_var}).rename(lam_var)
            lam_i = self.grid._ds[f"{lam_var}_i"]
            self.method = "xgcm"
        else:
            onedimensional_target = False
            if self.grid.axes['Z'].coords['center'] in lam.dims:
                lam_i = (
                    self.grid.interp(lam, "Z", boundary="extend")
                    .rename(f"{lam.name}_i")
                )
            else:
                lam_i = lam.broadcast_like(self.grid._ds[self.grid.axes['Z'].coords['center']])

        if lambda_name in ([] if self.lambdas("density") is None else self.lambdas("density")):
            hlamdot_transformed = []
            for component in ["heat", "salt"]:
                if hlamdot[component] is not None:
                    bin_bounds = bins.values if isinstance(bins, xr.DataArray) else bins
                    # xhistogram cases
                    if (((self.method == "default") and integrate) or
                        (self.method == "xhistogram")):
                        hlamdot_transformed_component = histogram(
                            lam,
                            bins=[bin_bounds],
                            dim=dim,
                            weights=hlamdot[component].fillna(0.),
                            bin_dim_suffix="",
                            # TEMPORARY FIX FOR https://github.com/xgcm/xhistogram/issues/16
                            block_size=None
                        ).rename({lam.name:f"{lam.name}_l_target"})
                    # xgcm cases
                    elif (((self.method == "default") and not integrate) or
                          (self.method == "xgcm")):
                        hlamdot_transformed_component = self.grid.transform(
                            hlamdot[component].fillna(0.),
                            "Z",
                            target=bin_bounds,
                            target_data=lam_i,
                            method="conservative",
                        ).fillna(0.).rename({lam_i.name: f"{lam.name}_l_target"})
                        if integrate:
                            hlamdot_transformed_component = hlamdot_transformed_component.sum(
                                [self.grid.axes['X'].coords['center'],
                                 self.grid.axes['Y'].coords['center']]
                            )
                    hlamdot_transformed.append(
                        (hlamdot_transformed_component / np.diff(bin_bounds)).rename(f"{term}_{component}")
                    )
            hlamdot_transformed = xr.merge(hlamdot_transformed)
        else:
            bin_bounds = bins.values if isinstance(bins, xr.DataArray) else bins
            if (((self.method == "default") and integrate) or
                (self.method == "xhistogram")):
                hlamdot_transformed = histogram(
                    lam,
                    bins=[bin_bounds],
                    dim=dim,
                    weights=hlamdot.fillna(0.),
                    bin_dim_suffix="",
                    # TEMPORARY FIX FOR https://github.com/xgcm/xhistogram/issues/16
                    block_size=None
                ).rename({lam.name:f"{lam.name}_l_target"})
            elif (((self.method == "default") and not integrate) or
                  (self.method == "xgcm")):
                hlamdot_transformed = self.grid.transform(
                    hlamdot.fillna(0.),
                    "Z",
                    target=bin_bounds,
                    target_data=lam_i,
                    method="conservative"
                ).fillna(0.).rename({lam_i.name: f"{lam.name}_l_target"})
                if integrate:
                    hlamdot_transformed = hlamdot_transformed.sum(
                        [self.grid.axes['X'].coords['center'],
                         self.grid.axes['Y'].coords['center']]
                    )
            hlamdot_transformed = (
                hlamdot_transformed / np.diff(bin_bounds)
            ).rename(f"{term}")
        return hlamdot_transformed

    def transformations_from_hlamdot(self, lambda_name, term=None, bins=None, mask=None, integrate=True):
        """
        Lazily compute extensive tendencies, transform them into lambda space
        along the vertical ("Z") dimension, and optionally integrate along the
        horizontal dimensions ("X", "Y").

        Parameters
        ----------
        lambda_name : str
            Specifies lambda (e.g., 'heat', 'salt', 'sigma0', etc.). Use `lambdas()` for a list of available lambdas.
        term : str or list of str (default: None)
            Specifies process term (e.g., 'boundary_forcing', 'vertical_diffusion', etc.). Use `self.available_processes()`
            to list all terms. If None, defaults to list of all terms.
        bins : array like (default None)
            np.array with lambda values specifying the edges for each bin. If not specified, array will be automatically
            derived from the scalar field of lambda (e.g., temperature).
        mask : xr.DataArray (default: None)
            Boolean region mask (with same X and Y grid dimensions as `grid._ds` variables).
            If None, generate an all-True mask for domain-wide calculations.
        integrate : bool (default: True)
            Whether to integrate the result over ("X", "Y") dimensions.
        
        Returns
        -------
        transformations : xarray.Dataset
            Dataset containing components of water mass transformations, possibly grouped as
            specified by the arguments.

        See also
        --------
        self.integrate_transformations, self.map_transformations
        """
        
        if isinstance(term, str):
            terms = [term]
        elif isinstance(term, list):
            terms = term
        elif term is None:
            terms = self.available_processes()
        else:
            return
            
        wmts = []
        for term in terms:
            transformed_hlamdot = self.transform_hlamdot_term(lambda_name, term=term, bins=bins, mask=mask, integrate=integrate)
            if transformed_hlamdot is not None:
                wmts.append(-transformed_hlamdot)
            else:
                print(f"Process '{term}' for component {lambda_name} is unavailable.")
        return xr.merge(wmts)

    ### Helper function to groups terms based on density components (sum_components)
    ### and physical processes (group_processes)
    # Calculate the sum of grouped terms
    def _sum_terms(self, ds_terms, newterm, terms):
        das = []
        if isinstance(ds_terms, xr.DataArray) and isinstance(terms, str):
            if terms == ds_terms.name:
                das.append(ds_terms[term])
        elif isinstance(ds_terms, xr.Dataset):
            for term in terms:
                if term in ds_terms:
                    das.append(ds_terms[term])
        if len(das):
            ds_terms[newterm] = sum(das)

    def _group_processes(self, hlamdot):
        if hlamdot is None:
            return
        for c in hlamdot.coords:
            lambda_key = self.get_lambda_key(c.split("_")[0])
            if (lambda_key is not None):
                if lambda_key == "density":
                    suffixes = ["_heat", "_salt", ""]
                    budget = self.xbudget_dict["heat"]
                else:
                    suffixes = [""]
                    budget = self.xbudget_dict[lambda_key]
                for suffix in suffixes:
                    if 'lhs' in budget:
                        self._sum_terms(
                            hlamdot,
                            f"kinematic_transformation{suffix}",
                            [f"{term}{suffix}" for term in budget['lhs'].keys()]
                        )
                    if 'rhs' in budget:
                        self._sum_terms(
                            hlamdot,
                            f"material_transformation{suffix}",
                            [f"{term}{suffix}" for term in budget['rhs'].keys()]
                        )
        return hlamdot

    def _sum_components(self, hlamdot, group_processes = False):
        if hlamdot is None:
            return
        
        for proc in self.available_processes():
            proc_list = [
                f"{proc}{suffix}"
                for suffix in ["_heat", "_salt"]
            ]
            self._sum_terms(
                hlamdot,
                proc,
                proc_list
            )
        if group_processes:
            for proc in [
                "kinematic_transformation",
                "material_transformation"
            ]:
                self._sum_terms(
                    hlamdot,
                    proc,
                    [f"{proc}{suffix}" for suffix in ["_heat", "_salt"]]
                )
        return hlamdot

    def map_transformations(
        self,
        lambda_name,
        term=None,
        group_processes=False,
        sum_components=True,
        bins=None,
        mask=None
        ):
        """
        Lazily compute column-wise water mass transformations.

        Parameters
        ----------
        lambda_name : str
            Specifies lambda (e.g., 'heat', 'salt', 'sigma0', etc.). Use `lambdas()` for a list of available lambdas.
        term : str or list of str (default: None)
            Specifies process term (e.g., 'boundary_forcing', 'vertical_diffusion', etc.). Use `self.available_processes()`
            to list all terms. If None, defaults to list of all terms.
        group_processes: bool (default: False)
            Specify whether process terms are summed to categories forcing and diffusion.
        sum_components : bool (default: True)
            Specify whether heat and salt tendencies are summed together (True) or kept separated (False).
        bins : array like (default None)
            np.array with lambda values specifying the edges for each bin. If not specified, array will be automatically
            derived from the scalar field of lambda (e.g., temperature).
        mask : xr.DataArray (default: None)
            Boolean region mask (with same X and Y grid dimensions as `grid._ds` variables).
            If None, generate an all-True mask for domain-wide calculations.
        
        Returns
        -------
        transformations : xarray.Dataset
            Dataset containing components of water mass transformations, possibly grouped as
            specified by the arguments.

        Example
        --------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt_rates = wmt.map_transformations("heat", term=["diffusion"])
        
        See also
        --------
        self.transformations_from_hlamdot, self.integrate_transformations
        """
        
        # call the base function
        transformations = self.transformations_from_hlamdot(
            lambda_name,
            term=term,
            bins=bins,
            mask=mask,
            integrate=False,
        )

        # process this function arguments
        if sum_components:
            transformations = self._sum_components(transformations, group_processes=group_processes)
        if group_processes:
            transformations = self._group_processes(transformations)
            
        return transformations

    def integrate_transformations(
        self,
        lambda_name,
        term=None,
        group_processes=False,
        sum_components=True,
        bins=None,
        mask=None,
        ):
        """
        Lazily compute horizontally-integrated water mass transformations.

        Parameters
        ----------
        lambda_name : str
            Specifies lambda (e.g., 'heat', 'salt', 'sigma0', etc.). Use `lambdas()` for a list of available lambdas.
        term : str or list of str (default: None)
            Specifies process term (e.g., 'boundary_forcing', 'vertical_diffusion', etc.). Use `self.available_processes()`
            to list all terms. If None, defaults to list of all terms.
        group_processes: bool (default: False)
            Specify whether process terms are summed to categories forcing and diffusion.
        sum_components : bool (default: True)
            Specify whether heat and salt tendencies are summed together (True) or kept separated (False).
        bins : array like (default None)
            np.array with lambda values specifying the edges for each bin. If not specified, array will be automatically
            derived from the scalar field of lambda (e.g., temperature).
        mask : xr.DataArray (default: None)
            Boolean region mask (with same X and Y grid dimensions as `grid._ds` variables).
            If None, generate an all-True mask for domain-wide calculations.
        
        Returns
        -------
        transformations : xarray.Dataset
            Dataset containing components of water mass transformations, possibly grouped as
            specified by the arguments.

        Example
        --------
        >>> wmt = WaterMassTransformations(grid, xbudget_dict)
        >>> wmt_rates = wmt.map_transformations("heat", term=["diffusion"])

        See also
        --------
        self.transformations_from_hlamdot, self.map_transformations
        """
        
        # call the base function
        transformations = self.transformations_from_hlamdot(
            lambda_name,
            term=term,
            bins=bins,
            mask=mask,
            integrate=True,
        )

        # process this function arguments
        if sum_components:
            transformations = self._sum_components(transformations, group_processes=group_processes)
        if group_processes:
            transformations = self._group_processes(transformations)
            
        return transformations
